/**
 * @file help.h
 * @author Joe Wingbermuehle
 * @date 2004-2006
 *
 * @brief Header for the help functions.
 *
 */

#ifndef HELP_H
#define HELP_H

/** Display program name, version, and compiled options . */
void DisplayAbout(void);

/** Display compiled options. */
void DisplayCompileOptions(void);

/** Display all help. */
void DisplayHelp(void);

#endif /* HELP_H */

