/**
 * @file grab.h
 * @author Joe Wingbermuehle
 * @date 2013
 *
 * @brief Header file for managing server grabs.
 *
 */

#ifndef GRAB_H
#define GRAB_H

/** Grab the server and sync. */
void GrabServer(void);

/** Ungrab the server. */
void UngrabServer(void);

#endif /* GRAB_H */

